"""Manage user case files.


"""
import importlib.resources
import inspect
import os
import pkgutil
import shutil
from itertools import chain
from pathlib import Path
from socket import gethostname

from fluidsim.base.output.base import OutputBase
from snek5000 import get_asset, logger, mpi


class Output(OutputBase):
    """Container and methods for getting paths of and copying case files.

    A path object :code:`self.root` points to the directory containing the
    case files.

    """

    @property
    def excludes(self):
        """Prefixes and suffixes of files which should be excluded from being
        copied."""
        return {"prefix": "__", "suffix": (".vimrc", ".tar.gz", ".o", ".py")}

    @property
    def makefile_usr_sources(self):
        """
        Sources for inclusion to makefile_usr.inc
        Dict[directory]  -> list of source files
        """
        return {
            # "source_directory": [
            #    (src1a.f, src1b.f)
            #    (src2a.f, src2b.f, src2c.f), ...
            # ]
        }

    @property
    def makefile_usr_obj(self):
        """Object files to be included in compilation. Should be exported as USR
        environment variable.

        """
        makefile_usr_obj = [
            sources[0].replace(".f", ".o")
            for sources in chain.from_iterable(self.makefile_usr_sources.values())
        ]
        return makefile_usr_obj

    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the ParamContainer info_solver."""
        OutputBase._complete_info_solver(info_solver)

        classes = info_solver.classes.Output.classes

        classes.PrintStdOut.module_name = "snek5000.output.print_stdout"
        classes.PrintStdOut.class_name = "PrintStdOut"
        classes.PhysFields.module_name = "snek5000.output.phys_fields"
        classes.PhysFields.class_name = "PhysFields"

    @staticmethod
    def _complete_params_with_default(params, info_solver):
        """This static method is used to complete the *params* container.
        """
        # Bare minimum
        attribs = {
            "ONLINE_PLOT_OK": True,
            "period_refresh_plots": 1,
            "HAS_TO_SAVE": True,
            "sub_directory": "",
        }
        params._set_child("output", attribs=attribs)

    @classmethod
    def get_root(cls):
        """Get the path to the current package."""
        # Better than
        # >>> root = Path(__file__).parent?
        # >>> with importlib.resources.path(__name__, "__init__.py") as f:
        # ...     return f.parent
        root = Path(inspect.getmodule(cls).__file__).parent
        return root

    @classmethod
    def get_configfile(cls, host=None):
        """Get path of the Snakemake configuration file for the current machine.
        All configuration files are stored under ``etc`` sub-package.

        """
        if not host:
            host = os.getenv(
                "SNIC_RESOURCE", os.getenv("GITHUB_WORKFLOW", gethostname())
            )
        root = cls.get_root()
        configfile = root / "etc" / f"{host}.yml"

        if not configfile.exists():
            logger.warning(
                "Expected a configuration file describing compilers and flags: "
                f"{configfile}"
            )
            configfile = Path(get_asset("default_configfile.yml"))
            logger.info(f"Using default configuration instead: {configfile}")

        return configfile

    def __init__(self, sim=None):
        self.sim = sim
        try:
            self.name_solver = sim.info.solver.short_name
        except AttributeError:
            pass

        self.root = self.get_root()
        # Check configfile early
        self.get_configfile()

        if sim:
            # Same as package name __name__
            self.name_pkg = sim.info.solver.classes.Output.module_name.split(".")[0]
            super().__init__(sim)

    def _get_resources(self, name_pkg=None):
        """Get a generator of resources (files) in a package, excluding
        directories (subpackages).

        :returns: generator

        """
        excludes = self.excludes
        if not name_pkg:
            name_pkg = self.name_pkg
        try:
            contents_pkg = importlib.resources.contents(name_pkg)
        except TypeError:
            return ()
        return (
            f
            for f in contents_pkg
            if (
                importlib.resources.is_resource(name_pkg, f)
                and not any(f.startswith(ext) for ext in excludes["prefix"])
                and not any(f.endswith(ext) for ext in excludes["suffix"])
            )
        )

    def _get_subpackages(self):
        """Get a dictionary of subpackages with values generated by
        :func:`get_resources`.

        :returns: dict

        """
        root = self.root
        name_pkg = self.name_pkg
        return {
            subpkg.name.replace(f"{root.name}.", ""): self._get_resources(subpkg.name)
            for subpkg in pkgutil.walk_packages([str(root)], prefix=f"{name_pkg}.")
        }

    def get_paths(self):
        """Get a list of paths to all case files.

        :returns: list

        """
        paths = []

        # abl.usr -> /path/to/abl/abl.usr
        paths += [self.root / resource for resource in self._get_resources()]

        for subpkg, resources in self._get_subpackages().items():
            # toolbox -> /path/to/abl/toolbox
            subpkg_root = self.root / subpkg.replace(".", os.sep)
            # main.f -> /path/to/abl/toolbox/main.f
            paths += [subpkg_root / resource for resource in resources]
        return paths

    def copy(self, new_dir, force=False):
        """Copy case files to a new directory. The directory does not have to be present.

        :param new_dir: A str or Path-like instance pointing to the new directory.
        :param force: Force copy would overwrite if files already exist.

        """
        # Avoid race conditions! Should be only executed by rank 0.
        if mpi.rank != 0:
            return

        abs_paths = self.get_paths()
        subpackages = self._get_subpackages()
        root = self.root

        def conditional_ignore(src, names):
            """Ignore if not found in ``abs_paths``."""
            src = Path(src)
            include = abs_paths + [root / subpkg for subpkg in subpackages]
            exclude = tuple(
                name
                for name in names
                if not any((src / name) == path for path in include)
            )

            logger.debug(
                "".join(
                    (
                        f"- src: {src}",
                        "\n- include:\n",
                        "\n ".join(str(name) for name in include),
                        "\n- exclude:\n",
                        "\n ".join(exclude),
                        "\n----",
                    )
                )
            )
            return exclude

        new_root = Path(new_dir)
        try:
            logger.info("Copying with shutil.copytree ...")
            copytree_kwargs = dict(
                src=root, dst=new_root, symlinks=False, ignore=conditional_ignore
            )
            # Python 3.8+
            shutil.copytree(**copytree_kwargs, dirs_exist_ok=True)
        except (TypeError, shutil.Error):
            try:
                logger.warning(
                    "Python < 3.8: shutil.copytree may not proceed if directories exist."
                )
                # Hoping that new_root has not been created
                shutil.copytree(**copytree_kwargs)
            except FileExistsError as e:
                logger.warning(e)
                logger.info("Copying with shutil.copy2 ...")
                # Copy one by one from the scratch
                if not new_root.exists():
                    logger.debug(f"Creating {new_root} ...")
                    os.makedirs(new_root, exist_ok=True)

                for abs_path in abs_paths:
                    rel_path = abs_path.relative_to(root)
                    new_path = new_root / rel_path
                    if not new_path.parent.exists():
                        os.makedirs(new_path.parent)

                    logger.debug(f"Copying {new_path}")
                    if new_path.exists():
                        if force:
                            logger.warning(f"{new_path} would be overwritten ...")
                        else:
                            logger.warning(
                                f"{new_path} exists, skipping. Use force=True to overwrite."
                            )
                            continue

                    shutil.copy2(abs_path, new_path)
        finally:
            logger.info(f"Copied: {root} -> {new_root}")

    def write_box(self, template):
        """Write <case name>.box file from box.j2 template."""
        if mpi.rank == 0:
            box_file = self.sim.path_run / f"{self.name_pkg}.box"
            logger.info(f"Writing box file... {box_file}")
            with open(box_file, "w") as fp:
                self.sim.oper.write_box(
                    template, fp, comments=self.sim.params.short_name_type_run
                )

    def write_size(self, template):
        """Write SIZE file from SIZE.j2 template."""
        if mpi.rank == 0:
            size_file = self.sim.path_run / "SIZE"
            logger.info(f"Writing SIZE file... {size_file}")
            with open(size_file, "w") as fp:
                self.oper.write_size(
                    template, fp, comments=self.sim.params.short_name_type_run
                )

    def write_makefile_usr(self, template):
        """Write the makefile_usr.inc file which gets included in the main
        makefile.

        """

        paths_of_sources = []

        for path_dir, list_of_sources in self.makefile_usr_sources.items():
            for sources in list_of_sources:
                paths_of_sources.append([f"{path_dir}/{file}" for file in sources])

        if mpi.rank == 0 and paths_of_sources:
            makefile_usr = self.sim.path_run / "makefile_usr.inc"
            output = template.render(
                list_of_sources=paths_of_sources,
                comments=self.sim.params.short_name_type_run,
            )
            with open(makefile_usr, "w") as fp:
                fp.write(output)



def test_init(oper):
    print(oper.produce_str_describing_oper())
    print(oper.produce_long_str_describing_oper())


def test_properties(oper):
    assert oper.max_n_seq == 1024
    assert oper.max_n_loc == 171


def test_box_template(oper):
    from phill.templates import box

    oper.write_box(box, comments=__name__)


def test_size_template(sim):
    from phill.templates import size

    sim.oper.write_size(size, comments=__name__)



"""Calculate physical constants"""
import math


def coriolis_freq(lat):
    """Compute coriolis frequency.

    https://en.wikipedia.org/wiki/Earth%27s_rotation#Angular_speed
    omega = 7.2921150 ± 0.0000001×10−5 radians per second

    :param theta: Latitude in degrees.

    """
    omega = 7.2921150e-5
    theta = math.radians(lat)
    f = 2 * omega * math.sin(theta)
    return f



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

import os
import subprocess
import sys
from datetime import date
from pathlib import Path
from subprocess import PIPE

import breathe
import snek5000
from snek5000 import util


def root(module):
    return os.fspath(Path(module.__file__).parent)


sys.path.insert(0, root(breathe))

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, root(snek5000))

print("sys.path =\n   ", "\n    ".join(sys.path))

# -- Project information -----------------------------------------------------

project = "snek5000"
_today = date.today()
copyright = (
    f"2019 - {_today.year}, Ashwin Vishnu Mohanan. Published: {_today.isoformat()}"
)
author = "Ashwin Vishnu Mohanan"

version = ".".join(snek5000.__version__.split(".")[:3])
# The full version, including alpha/beta/rc tags
release = snek5000.__version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "recommonmark",
]

# Execute Doxygen
os.makedirs("_build/html/doxygen", exist_ok=True)

# Inspect whether to run doxygen or not
last_modified = util.last_modified("../lib").stat().st_mtime
timestamp = Path("_build/.doxygen_timestamp")
if timestamp.exists() and Path("_build/xml").exists():
    with open(timestamp) as fp:
        last_documented = float(fp.read())
    exec_doxygen = last_documented < last_modified
else:
    exec_doxygen = True

# Modify Doxygen configuration or not
#  modify_doxygen = any(os.getenv(env) for env in ("CI", "GITHUB_ACTIONS", "READTHEDOCS"))
modify_doxygen = False
if modify_doxygen:
    print("Disabling source browser... ", end="")

    # Disable source browser
    with open("Doxyfile", "rb") as doxyfile:
        doxy_cfg = [
            line for line in doxyfile.readlines() if b"SOURCE_BROWSER" not in line
        ]
    doxy_cfg = b"".join(doxy_cfg)
    # print(doxy_cfg.decode("utf8"))


try:
    if exec_doxygen:
        print("Executing Doxygen... ", end="")
        if modify_doxygen:
            # Pass configuration via stdin
            with subprocess.Popen(["doxygen", "-"], stdin=PIPE, stdout=PIPE) as proc:
                doxy_output = proc.communicate(input=doxy_cfg)[0]
        else:
            doxy_output = subprocess.check_output(["doxygen"])

        doxy_summary = doxy_output.decode("utf8").splitlines()[-2:]
        print("done:", *doxy_summary)
        with open(timestamp, "w") as fp:
            fp.write(str(last_modified))
    else:
        print(
            f"Using old Doxygen XML output... Remove {timestamp} to force doxygen build."
        )
except FileNotFoundError:
    print("Can not find doxygen to generate the documentation of the Fortran code.")
else:
    # -- Breathe configuration ---------------------------------------------------
    extensions.append("breathe")

    # breathe_default_members = ('members', 'undoc-members')

    # File types
    breathe_implementation_filename_extensions = [".md"]
    breathe_domain_by_extension = {"usr": "c", "inc": "c", "f": "c"}
    breathe_domain_by_file_pattern = {"SIZE": "f"}

    # Input sources
    breathe_projects = {"snek5000": "_build/xml/"}
    #  breathe_projects_source = {
    #      "abl": ("../src/abl", ["SIZE", "abl.usr"]),
    #  }
    breathe_default_project = "snek5000"

# ----------------------------------------------------------------------------

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
source_suffix = [".rst", ".md"]


# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

html_favicon = "_static/favicon.ico"

# -- Options for Intersphinx -------------------------------------------------

intersphinx_mapping = dict(
    python=("https://docs.python.org/3", None),
    nek=("https://nek5000.github.io/NekDoc", None),
    jinja2=("https://jinja.palletsprojects.com/en/2.10.x", None),
)

# -- Other options ------------------------------------------------------------
autosummary_generate = True

autodoc_default_options = {
    "members": True,
    #  'member-order': 'bysource',
    #  'special-members': '__init__',
    #  'undoc-members': True,
    #  'exclude-members': '__weakref__'
}
autodoc_mock_imports = ["IPython"]

todo_include_todos = True

napoleon_numpy_docstring = True



"""Miscellaneous utilities
==========================


"""
import os
import sys
from datetime import datetime
from functools import reduce
from pathlib import Path
from tarfile import TarFile
from zipfile import ZipFile

from fluiddyn import util

from .. import source_root
from ..log import logger


def isoformat(dt):
    """Returns timestamp in modified isoformat from a datetime object (-
    instead of :). Accurate upto seconds.

    .. note:: The modified ISO format is YYYY-MM-DDTHH-MM-SS

    """
    return dt.isoformat(timespec="seconds").replace(":", "-")


def timestamp(path):
    """Modification date of a file or directory as a timestamp.

    :returns str:

    """
    return isoformat(util.modification_date(path))


def modification_date(path):
    """Modification date of a file or directory. Returns empty string if does
    not exist.

    """
    if os.path.exists(path):
        t = os.path.getmtime(path)
        return isoformat(datetime.fromtimestamp(t))
    else:
        return ""


def now():
    """The current timestamp.

    :returns str:

    """
    return isoformat(datetime.now())


def zip_info(filename):
    """Contents of a zip file."""
    with ZipFile(filename) as file:
        # Print all contents
        file.printdir()


def tar_info(filename):
    """Contents of a tar file."""
    with TarFile(filename) as file:
        # Print all contents
        file.list()


def scantree(path):
    """Recursively yield DirEntry objects for given directory.

    Courtesy: https://stackoverflow.com/a/33135143

    :returns generator: A generator for ``DirEntry`` objects.

    """
    for entry in os.scandir(path):
        if entry.is_dir(follow_symlinks=False):
            yield from scantree(entry.path)
        else:
            yield entry


def last_modified(path):
    """Find the last modified file in a directory tree.

    :returns DirEntry:

    """
    return reduce(
        lambda x, y: x if x.stat().st_mtime > y.stat().st_mtime else y, scantree(path),
    )


def activate_paths():
    """Setup environment variables in preparation for Nek5000 build."""
    env_source_root = os.environ["NEK_SOURCE_ROOT"] = source_root()

    env_path = str(os.getenv("PATH"))
    if env_source_root not in env_path:
        os.environ["PATH"] = ":".join((env_source_root + "/bin", env_path))

    return env_source_root, env_path


def init_params(Class, isolated_unit=False):
    """Instantiate an isolated ``params`` for a specific class."""
    if hasattr(Class, "create_default_params") and not isolated_unit:
        params = Class.create_default_params()
    else:
        from ..params import Parameters

        params = Parameters(tag="params")
        Class._complete_params_with_default(params)

    return params


def docstring_params(Class, sections=False, indent_len=4):
    """Creates a parameter instance and generate formatted docs for it. The
    docs are defined by the ``params.<child>._set_doc`` method. Done only when
    ``sphinx`` is already imported.

    """
    if "sphinx" in sys.modules:
        params = init_params(Class, isolated_unit=True)
        doc = params._get_formatted_docs()
    else:
        doc = ""

    if not sections:
        lines = []
        indent = " " * indent_len
        prev_line = ""

        for line in doc.splitlines():
            if line.startswith("Documentation for"):
                lines.append(f"{indent}**{line.lstrip()}**")
            elif "Documentation for" in prev_line and any(
                line.startswith(underline) for underline in ("====", "----", "~~~~")
            ):
                continue
            else:
                lines.append(indent + line)

            prev_line = line

        doc = "\n".join((lines))

    return doc


def get_status(path):
    """Get status of a simulation run by verifying its contents.

    :returns tuple(int, str): status code and message

    """
    locks_dir = path / ".snakemake" / "locks"
    contents = os.listdir(path)

    if not (path / ".snakemake").exists():
        return (
            425,
            f"Too Early: Seems like snakemake was never executed: {path}",
        )
    elif locks_dir.exists():
        locks = tuple(locks_dir.iterdir())
        if locks:
            return (
                423,
                f"Locked: The path is currently locked by snakemake: {path}",
            )

    if not {"SIZE", "SESSION.NAME", "nek5000"}.issubset(contents):
        return (
            404,
            (
                "Not Found: SIZE and/or nek5000 and/or SESSION.NAME files are missing: "
                f"{path}: Contents: {contents}"
            ),
        )
    if not tuple(path.glob("rs6*")):
        return (404, f"Not Found: No restart files found: {path}")

    return (200, f"OK: All prerequisities satisfied to restart: {path}")


def prepare_for_restart(path, chkp_fnumber=1, verify_contents=True):
    """Takes a directory in which a simulation was executed.

    * If verify contents:
      - check if snakemake was ever executed or check if directory is locked by snakemake
      - ensures simulation files exist
      - ensures restart files exist
    * Reads params.xml if it exists, and if not falls back to .par file.
    * Modifies checkpoint parameters with appropriate ``chkp_fnumber`` to
      restart from KTH framework.

    """
    contents = os.listdir(path)
    path = Path(path)

    if verify_contents:
        status, msg = get_status(path)
        if status >= 400:
            logger.error(f"{status}: {msg}")
        else:
            logger.info(f"{status}: {msg}")

    # FIXME: make this generic for all possible solvers
    # Trying to read the par file
    solver_prefix = path.absolute().name.split("_")[0]
    try:
        from importlib import import_module

        Simul = import_module(f"{solver_prefix}.solver").Simul
    except ImportError:
        raise ImportError("Cannot import simulation class")

    if "params.xml" in contents:
        params = Simul.load_params_from_file(path_xml=str(path / "params.xml"))
    else:
        logger.warning(f"No params.xml found... Attempting to read {solver_prefix}.par")

        params = Simul.create_default_params()
        params.nek._read_par(path / f"{solver_prefix}.par")

    params.nek.chkpoint.chkp_fnumber = chkp_fnumber
    params.nek.chkpoint.read_chkpt = True

    return params



"""Nek5000 base solver
======================

A bare Nek5000 solver which does not rely on any user parameters.

"""
import math
from pathlib import Path
from warnings import warn

import numpy as np
from fluidsim.base.solvers.base import SimulBase
from fluidsim.base.solvers.info_base import create_info_simul

from .. import __version__, logger, mpi
from ..info import InfoSolverNek
from ..params import Parameters, create_params


class SimulNek(SimulBase):
    """Simulation class

    .. code-block:: python

       from snek5000.solvers.base import Simul
       params = Simul.create_default_params()
       sim = Simul(params)

    """

    InfoSolver = InfoSolverNek

    @classmethod
    def create_default_params(cls):
        """Generate default parameters. ``params.nek`` contains runtime
        parameters consumed by Nek5000.

        """
        cls.info_solver = cls.InfoSolver()
        cls.info_solver.complete_with_classes()
        return create_params(cls.info_solver)

    @classmethod
    def load_params_from_file(cls, path_xml=None, path_par=None):
        if not (path_xml or path_par):
            raise IOError("Either path to params.xml or case.par should be provided")

        params = Parameters(tag="params")
        if path_xml:
            params._load_from_xml_file(path_xml)
        else:
            logger.warn(
                "Loading from a par file will not have full details of the simulation"
            )
            params.nek._read_par(path_par)

        cls._set_internal_sections(params)
        return params

    @classmethod
    def _set_internal_sections(cls, params):
        try:
            info_solver = cls.info_solver
        except AttributeError:
            info_solver = cls.InfoSolver()

        for section in info_solver.par_sections:
            getattr(params.nek, section)._set_internal_attr("_user", False)

        for section in info_solver.par_sections_disabled:
            getattr(params.nek, section)._set_internal_attr("_enabled", False)

    @staticmethod
    def _complete_params_with_default(params):
        """A static method used to complete the *params* container.

        The sections are:

        * ``general`` (mandatory)
        * ``problemtype``
        * ``mesh``
        * ``velocity``
        * ``pressure`` (required for velocity)
        * ``temperature``
        * ``scalar%%``
        * ``cvode``

        When scalars are used, the keys of each scalar are defined under the section
        ``scalar%%`` varying between ``scalar01`` and ``scalar99``.

        """

        params._set_child("nek")
        params_nek = params.nek

        params._set_attribs(dict(NEW_DIR_RESULTS=True, short_name_type_run="run"))
        for section in ("general", "problemtype", "velocity", "pressure"):
            params_nek._set_child(section)

        for section in ("mesh", "temperature", "scalar01", "cvode"):
            params_nek._set_child(section)
            getattr(params_nek, section)._set_internal_attr("_enabled", False)

        SimulNek._set_internal_sections(params)

        params_nek.general._set_attribs(
            dict(
                start_from="",
                stop_at="numSteps",
                end_time=math.nan,
                num_steps=1,
                dt=math.nan,
                variable_dt=True,
                target_cfl=0.5,
                write_control="timeStep",
                write_interval=10,
                filtering=None,
                filter_cutoff_ratio=0.65,
                filter_weight=12.0,
                write_double_precision=True,
                dealiasing=True,
                time_stepper="BDF2",
                extrapolation="standard",
                opt_level=2,
                log_level=2,
                user_params={},
            )
        )

        params_nek.problemtype._set_attribs(
            dict(
                equation="incompNS",
                variable_properties=False,
                stress_formulation=False,
            )
        )
        common = dict(residual_tol=math.nan, residual_proj=False)
        params_nek.velocity._set_attribs(common)
        params_nek.pressure._set_attribs(common)
        params_nek.temperature._set_attribs(common)
        params_nek.scalar01._set_attribs(common)

        params_nek.velocity._set_attribs(dict(viscosity=math.nan, density=math.nan))
        params_nek.pressure._set_attrib("preconditioner", "semg_xxt")
        return params

    def __init__(self, params):
        np.seterr(all="warn")
        np.seterr(under="ignore")

        if (
            not hasattr(self, "info_solver")
            or self.info_solver.__class__ is not self.InfoSolver
        ):
            if hasattr(self, "info_solver"):
                warn(
                    "Creating a new info_solver instance "
                    f"due to type mismatch  {self.InfoSolver}"
                )
            self.info_solver = self.InfoSolver()
            self.info_solver.complete_with_classes()

        dict_classes = self.info_solver.import_classes()

        if not isinstance(params, Parameters):
            raise TypeError(
                f"params should be a Parameters instance, not {type(params)}"
            )

        self.params = params
        self.info = create_info_simul(self.info_solver, params)

        # initialize objects
        for cls_name, Class in dict_classes.items():
            setattr(self, cls_name.lower(), Class(self))

        if "Output" in dict_classes:
            # path_run would be initialized by the Output instance if available
            # See self.output._init_name_run()
            self.path_run = Path(self.output.path_run)
            self.output.copy(self.path_run)
        else:
            par_file = None
            self.path_run = None
            if mpi.rank == 0:
                logger.warning("No output class initialized!")

        _banner_length = 42
        if mpi.rank == 0:
            logger.info("*" * _banner_length)
            logger.info(f"solver: {self.__class__}")
            logger.info(f"path_run: {self.path_run}")
            logger.info("*" * _banner_length)
            if self.path_run:
                par_file = self.path_run / f"{self.output.name_pkg}.par"
                logger.info(f"Writing params files... {par_file}, params.xml")
                with open(par_file, "w") as fp:
                    params.nek._write_par(fp)

                params._save_as_xml(
                    self.path_run / "params.xml", f"eTurb version {__version__}"
                )


Simul = SimulNek



from pathlib import Path

import pytest


def pytest_addoption(parser):
    # https://pytest.readthedocs.io/en/latest/example/simple.html#control-skipping-of-tests-according-to-command-line-option
    parser.addoption(
        "--runslow", action="store_true", default=False, help="run slow tests"
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "slow: mark test as slow to run")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--runslow"):
        # --runslow given in cli: do not skip slow tests
        return
    skip_slow = pytest.mark.skip(reason="need --runslow option to run")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)


@pytest.fixture(scope="session")
def sim():
    from phill.solver import Simul

    params = Simul.create_default_params()
    params.output.sub_directory = "test"

    params.nek.general.stop_at = "numSteps"
    params.nek.general.num_steps = 9

    return Simul(params)


@pytest.fixture(scope="session")
def oper():
    from snek5000.operators import Operators as Class
    from snek5000.util import init_params

    params = init_params(Class)
    params.oper.nx = params.oper.ny = params.oper.nz = 9
    params.oper.nproc_min = 6

    return Class(params=params)


@pytest.fixture(scope="session")
def sim_data(tmpdir_factory):
    """Generate fake simulation data."""
    files = """box.tmp
makefile
makefile_usr.inc
nek5000
params.xml
phill0.f00001
phill.box
phill.f
phill.log
phill.ma2
phill.par
phill.re2
phill.usr
rs6phill0.f00001
rs6phill0.f00002
rs6phill0.f00003
SESSION.NAME
SIZE
Snakefile""".splitlines()

    path_run = Path(tmpdir_factory.mktemp("sim_data"))
    for f in files:
        (path_run / f).touch()

    return path_run



"""Paraview interface for Nek5000 files.

.. todo:: Load field files and make various plots of statistics

"""
from ..log import logger

try:
    import paraview.simple as pvs
    from paraview.simple import VisItNek5000Reader
except ImportError as e:
    logger.warning(e)

#  from vtktools import vtkio


class PhysFields:
    """Class for loading, plotting simulation files.

    :param sim: A simulation instance derived from
                :class:`snek5000.solvers.base.SimulNek`
    :param path: :class:`pathlib.Path` Path where simulation state files are
                 found.


    """

    def __init__(self, sim=None, path=None):
        self.sim = sim
        if not path and sim:
            self.path_run = sim.output.path_run
        else:
            self.path_run = path

    def read(self):
        path = self.path_run
        if path.is_dir():
            path = next(path.glob("*.nek5000"))

        # TODO: needed or not?
        # pvs._DisableFirstRenderCameraReset()
        self.reader = reader = VisItNek5000Reader(FileName=str(path))
        if not reader:
            raise IOError("Error while loading file " + path)

        reader.Meshes = ["mesh"]
        # get animation scene
        self.animationScene = pvs.GetAnimationScene()

        # get the time-keeper
        self.timeKeeper = pvs.GetTimeKeeper()

    def set_point_arrays(self, arrays=("velocity",)):
        self.reader.PointArrays = list(arrays)

    def contourf_paraview(self, array):
        assert array in self.reader.PointArrays
        renderView = pvs.GetActiveViewOrCreate("RenderView")

        # get color transfer function/color map for 'array'
        LUT = pvs.GetOpacityTransferFunction(array)

        # get color legend/bar for LUT in view renderView
        LUTColorBar = pvs.GetScalarBar(LUT, renderView)
        return renderView, LUTColorBar

    def set_time(self, time):
        self.animationScene.AnimationTime = time
        self.timeKeeper.Time = time



"""Snakemake interface
======================

"""
from snakemake import snakemake
from snakemake.executors import change_working_directory as change_dir


class Make:
    """Snakemake interface for the solvers.

    :param sim: A simulation instance

    """

    def __init__(self, sim):
        self.sim = sim
        self.path_run = sim.output.path_run
        try:
            self.file = next(f for f in sim.output.get_paths() if f.name == "Snakefile")
        except AttributeError:
            raise AttributeError("Unable to get path of Snakefile via Output class.")

        self.log_handler = []

    def list(self):
        """List rules."""
        with change_dir(self.path_run):
            return snakemake(self.file, listrules=True, log_handler=self.log_handler)

    def exec(self, rules=("run",), dryrun=False):
        """Execute snakemake rules in sequence.

        :returns: True if workflow execution was successful.

        """
        with change_dir(self.path_run):
            return snakemake(
                self.file, targets=rules, dryrun=dryrun, log_handler=self.log_handler
            )



"""Load and parse stdout from Nek5000.

"""
import re
from pathlib import Path

import numpy as np
from snek5000.log import logger
from snek5000.solvers.base import Simul

# Uses awkupy
# %load_ext iawk


class PrintStdOut:
    """Parse standard output log files."""

    _tag = "print_stdout"

    def __init__(self, output=None, file=None):
        self.output = output
        self._file = file

    @property
    def file(self):
        output = self.output
        if output and not self._file:
            logger.info("Searching for a log file...")
            path_run = Path(output.path_run)
            logfiles = sorted(path_run.glob("*.log"))
            if logfiles:
                self._file = logfiles[-1]
            else:
                raise FileNotFoundError(
                    f"Cannot find a .log to parse in {path_run}: {logfiles}"
                )

        return self._file

    @file.setter
    def file(self, path_log_file):
        self._file = path_log_file

    @property
    def text(self):
        with open(self.file) as fp:
            return fp.read()

    @property
    def path_run(self):
        """Parse path_run from log file"""
        par_file = re.findall("Reading.*par", self.text)[-1].split()[-1]
        path = Path(par_file).parent
        return path

    @property
    def params(self):
        return Simul.load_params_from_file(path_xml=self.path_run / "params.xml")

    @property
    def text_steps(self):
        """Parse text starting with Step

        https://regex101.com/r/enFOAg/1
        """
        pattern = re.compile(
            r"""^Step          # For all lines starting with Step
                \s*            # Followed by some whitespaces
                (\d+)          # 0: Capture timestep which are integers
                .*             # Followed by some characters until
                t=\s*          # t=spaces
                (\S+)          # 1: Capture t: any non-whitespace char
                ,\s*           # comma and spaces
                DT=\s*         # DT=spaces
                (\S+)          # 2: Capture DT: any non-whitespace char
                ,\s*           # comma and spaces
                C=\s*          # C=spaces
                (.+)           # 3: Capture C: any char
    """,
            re.VERBOSE | re.MULTILINE,
        )
        return pattern.findall(self.text)

    @property
    def dt(self):
        """Extract time step dt over the course of a simulation."""
        # Alternate implementations
        # 1. awkup
        # dts = %awk -F, -e '/Step/{print $3}' {self.file}
        #
        # 2. Basic regex
        # steps = re.findall("Step.*DT.*", self.text)
        # try:
        #     dt_strings = [step.split(',')[2] for step in steps if step]
        #     dts = [float(s.split("= ")[1]) for s in dt_strings if s]
        # except IndexError:
        #     print(steps)
        dt = [float(field[2]) for field in self.text_steps]
        return np.array(dt)



"""IPython magic extensions
===========================

Usage example in an IPython console:

.. code-block:: ipython

   In [1]: %load_ext snek5000.magic
   WARNING  Activating this magic involves dirty monkey-patching of fluidsim.magic

   In [2]: %snek5000 abl
   INFO     Reading baseline parameters from /home/avmo/src/exabl/snek5000/src/abl/abl.par
   Created Simul class and default parameters for abl -> Simul, params

   In [3]: sim = Simul(params)

"""
import pkgutil
from importlib import import_module

from fluidsim import magic as fluidsim_magic
from IPython.core import magic_arguments
from IPython.core.magic import line_magic, magics_class

from .log import logger


def available_solver_keys(package):
    package = import_module(package)
    modules = [
        mod.name for mod in pkgutil.walk_packages(package.__path__) if not mod.ispkg
    ]
    return modules


def import_simul_class_from_key(key, package):
    solver = import_module(f"{package}.{key}")
    return solver.Simul


# Monkey patch
fluidsim_magic.available_solver_keys = available_solver_keys
fluidsim_magic.import_simul_class_from_key = import_simul_class_from_key


@magics_class
class EturbMagics(fluidsim_magic.FluidsimMagics):
    @magic_arguments.magic_arguments()
    @magic_arguments.argument("solver", nargs="?", default="")
    @magic_arguments.argument("-f", "--force-overwrite", action="store_true")
    @line_magic
    def snek5000(self, line):
        super().fluidsim(line)


def load_ipython_extension(ipython):
    """Load the extension in IPython."""
    logger.warning(
        "Activating this magic involves dirty monkey-patching of fluidsim.magic"
    )
    magics = EturbMagics(ipython, "snek5000.solvers")
    ipython.register_magics(magics)



"""
A Pythonic frontend to Nek5000.

**Sub-packages**

.. autosummary::
   :toctree:

   output
   solvers

**Modules**

.. autosummary::
   :toctree:

   assets
   info
   log
   magic
   make
   operators
   params
   util

"""
import importlib.resources
import os

from fluiddyn.util import mpi  # noqa: F401

from ._version import __version__  # noqa: F401
from .log import logger  # noqa: F401


def source_root():
    """Path of Nek5000 source code."""
    if os.getenv("SOURCE_ROOT"):
        logger.warning(
            "The environment variable "
            "SOURCE_ROOT is deprecated in v19, use NEK_SOURCE_ROOT instead."
        )

    with importlib.resources.path(__name__, "__init__.py") as f:
        root = f.parent

    nek5000 = os.path.expandvars(
        os.path.expanduser(
            os.getenv("NEK_SOURCE_ROOT", str((root / "../../lib/Nek5000").absolute()))
        )
    )
    if not os.path.exists(nek5000):
        logger.error("NEK_SOURCE_ROOT: " + nek5000)
        raise IOError(
            "Cannot find Nek5000 source code. Use environment variable "
            "NEK_SOURCE_ROOT to point to Nek5000 top level directory."
        )
    return nek5000


def get_asset(asset_name):
    """Fetches path of an asset from subpackage ``snek5000.assets``."""
    asset = next(importlib.resources.path("snek5000.assets", asset_name).gen)
    return asset



try:
    from fluiddyn.clusters.snic import ClusterSNIC as Base, _venv

    class Cluster(Base):
        default_project = "snic2020-1-31"
        cmd_run = "_delete"
        cmd_run_interactive = "_delete"

        def __init__(self):
            super().__init__()
            self.commands_setting_env = [
                "ml purge",
                "ml restore",
                f"source activate {_venv}",
                'export LD_LIBRARY_PATH="$LD_LIBRARY_PATH":"$LIBRARY_PATH"',
                "cd ..",
                "source activate.sh",
                "cd bin/",
            ]

        def _create_txt_launching_script(self, **kwargs):
            txt = super()._create_txt_launching_script(**kwargs)
            return "\n".join(
                [line for line in txt.splitlines() if not line.startswith("_delete")]
            )


except ImportError as e:
    from .log import logger

    logger.warning(str(e))
    logger.info("Using local machine as a cluster")
    from fluiddyn.clusters.local import ClusterLocal as Cluster  # noqa



"""Operators
============

Information regarding mesh, mathematical operators, and domain decomposition.

.. todo:: Move length to abl.par file and extrude it to allow double precision
          values.

.. note:: We deliberately try not to use the variable names used in Nek5000, as
          those are ambiguously named.

"""
import inspect
import math
import sys
from math import pi

from .util import docstring_params


def _str_len(length):
    """Generates a special string if length is multiple of pi. If not provide 3
    decimal places past floating point.

    """
    if (length / pi).is_integer():
        str_len = repr(int(length / pi)) + "pi"
    else:
        str_len = f"{length:.3f}".rstrip("0")

    return str_len


def next_power(value, base=2):
    """Compute the perfect power of ``base`` greater than or equal to ``value``.

    :param value:
    :type value: int or float

    :param int base:

    :return int:

    """
    exponent = math.ceil(math.log(value, base))
    return base ** exponent


class Operators:
    """Container for parameters and writing box_ and SIZE_ files.

    .. _box: https://nek5000.github.io/NekDoc/appendix.html?\
#generating-a-mesh-with-genbox
    .. _SIZE: https://nek5000.github.io/NekDoc/problem_setup/case_files.html\
#size

    .. note:: Some values are not available as parameters and instead
              automatically computed for generating the SIZE file.

    ==============  ======================== ==================================
    SIZE            `properties`             Comment
    ==============  ======================== ==================================
    ``lelg``        :any:`max_n_seq`         Max. number of elements globally
    ``lelt``        :any:`max_n_loc`         | Max. number of elements per
                                               processor (should be not smaller
                                             | than ``lelg/lpmin``, i.e.

    ``lelx``        :any:`max_nx`            | **Automatically computed**. Max.
                                               number of elements along x
                                             | direction for the global tensor
                                               product solver / dimensions.

    ``lely``        :any:`max_ny`            Same as above for ``y`` direction.
    ``lelz``        :any:`max_nz`            Same as above for ``z`` direction.

    ``lbelt``       :any:`order_mhd`         | **Automatically computed** as
                                             |  ``lelt`` if ``"MHD" in
                                               params.nek.problem_type.equation``.
    ``lpelt``       :any:`order_linear`      | **Automatically computed** as
                                             |  ``lelt`` if ``"linear" in
                                               params.nek.problem_type.equation``.
    ``lcvelt``      :any:`order_cvode`       | **Automatically computed** as
                                             |  ``lelt`` if
                                               ``params.nek.cvode._enabled is
                                               True``

    ``lx1m``        :any:`order_mesh_solver` | p-order for mesh solver.
                                               **Automatically computed** based
                                             | ``params.nek.general.stress\
_formulation`` and whether
                                             | Arbitrary Lagrangian-Eulerian
                                               (ALE) methods are used or not.
    ==============  ======================== ==================================

    """

    @staticmethod
    def _complete_params_with_default(params):
        """This static method is used to complete the *params* container.
        """
        attribs = {
            "nx": 8,
            "ny": 8,
            "nz": 8,
            "origin_x": 0.0,
            "origin_y": 0.0,
            "origin_z": 0.0,
            "ratio_x": 1.0,
            "ratio_y": 1.0,
            "ratio_z": 1.0,
            "Lx": 2 * pi,
            "Ly": 2 * pi,
            "Lz": 2 * pi,
            "boundary": ["P", "P", "W", "W", "P", "P"],
            "dim": 3,
            "nproc_min": 4,
            "nproc_max": 32,
            "scalars": 1,
        }
        params._set_child("oper", attribs=attribs)
        params.oper._set_doc(
            """

The following table matches counterpart of mandatory ``SIZE`` variables.

==========  ===================   =============================================
SIZE        params.oper           Comment
==========  ===================   =============================================
``ldim``    ``dim``               Domain dimensions (2 or 3)

``lpmin``   ``nproc_min``         Min MPI ranks
``lpmax``   ``nproc_max``         Max MPI ranks
``ldimt``   ``scalars``           Number of auxilary fields (minimum 1).

==========  ===================   =============================================

"""
        )

        attribs = {
            option: 1
            for option in (
                "hist",
                "obj",
                "perturb",
                "scalars_cons",
                "scalars_proj",
                "sessions",
            )
        }
        attribs["dim_proj"] = 20
        attribs["dim_krylov"] = 30
        attribs["order_time"] = 2

        params.oper._set_child("max", attribs=attribs)
        params.oper.max._set_doc(
            """
The following table matches counterpart of optional ``SIZE`` variables. These
refer to upper bound number of `something`. The parameters are considered
"optional" and would be ignored with the default values.

==============  ===================   =========================================
SIZE            params.oper.max       Comment
==============  ===================   =========================================
``mxprev``      ``dim_proj``          Max. dimension of projection space
``lgmres``      ``dim_krylov``        Max. dimension of Krylov space for GMRES
``lhis``        ``hist``              Max. number of history (i.e. monitoring)
                                      points.

``maxobj``      ``obj``               Max. number of objects?
``lorder``      ``order_time``        Max. temporal order (minimum: 2)
``lpert``       ``perturb``           Max. number of perturbations
``toteq``       ``scalars_cons``      Max. conserved scalars
``ldimt_proj``  ``scalars_proj``      Max. scalars for residual projection
``nsessmax``    ``sessions``          Max. sessions to NEKNEK

==============  ===================   =========================================

"""
        )

        attribs = {
            "order": 6,
            "order_out": 6,
            "coef_dealiasing": 2.0 / 3,
            "staggered": True,
        }
        params.oper._set_child("elem", attribs=attribs)
        params.oper.elem._set_doc(
            r"""
The following table relate to element configuration for certain operations.
The parameters are considered "optional" (except for ``lx1`` / ``order`` and
``lxd`` / ``coef_dealiasing`` which are mandatory) and would be ignored with
the default values.

==========      ===================   =========================================
SIZE            params.oper.elem      Comment
==========      ===================   =========================================
``lxd``         ``coef_dealiasing``   | p-order [#f1]_ for over-integration.
                                        **Automatically computed** from float
                                      | ``coef_dealiasing``. See
                                        :any:`order_dealiasing`

``lx1``         ``order``             p-order (avoid uneven and values <6).

``lxo``         ``order_out``         Max. p-order on output (should be ``>=
                                      order``)

``lx2``         ``staggered``         | p-order for pressure. **Automatically
                                        computed** from boolean
                                      | ``staggered`` (`True` implies
                                        :math:`\mathbb{P}_N - \mathbb{P}_{N-2}`
                                        and `False` implies
                                      | :math:`\mathbb{P}_N
                                        - \mathbb{P}_{N}` or a collocated
                                        grid). See :any:`order_pressure`

==========      ===================   =========================================

.. [#f1] Polynomial order which means the number of GLL / GL points per element.

"""
        )
        attribs = {"fast_diag": False}
        params.oper._set_child("misc", attribs=attribs)
        params.oper.misc._set_doc(
            r"""
Other miscellaneous parameters:

==========      ===================   =========================================
SIZE            params.oper.misc      Comment
==========      ===================   =========================================
``lfdm``        ``fast_diag``         | Equals to True for global tensor
                                        product solver (that uses fast
                                      | diagonalization method). ``False``
                                        otherwise.
==========      ===================   =========================================

"""
        )

    def __init__(self, sim=None, params=None):
        self.params = sim.params if sim else params
        self.axes = ("x", "y", "z")

    @property
    def max_n_seq(self):
        """Equivalent to ``lelg``."""
        params = self.params
        return next_power(params.oper.nx * params.oper.ny * params.oper.nz)

    @property
    def max_n_loc(self):
        """Equivalent to ``lelt``. The next integer greater than or equals
        ``max_n_seq / params.oper.nproc_min``.
        """
        return math.ceil(self.max_n_seq / self.params.oper.nproc_min)

    @property
    def max_nx(self):
        """Equivalent to ``lelx``."""
        return next_power(self.params.oper.nx)

    @property
    def max_ny(self):
        """Equivalent to ``lely``."""
        return next_power(self.params.oper.ny)

    @property
    def max_nz(self):
        """Equivalent to ``lelz``."""
        return next_power(self.params.oper.nz)

    @property
    def nb_fields(self):
        """Used in .box file."""
        return self.params.oper.scalars + 1

    @property
    def order(self):
        """Equivalent to ``lx1``."""
        return self.params.oper.elem.order

    @property
    def order_dealiasing(self):
        """Equivalent to ``lxd``."""
        elem = self.params.oper.elem
        return math.ceil(elem.order / elem.coef_dealiasing)

    @property
    def order_pressure(self):
        """Equivalent to ``lx2``."""
        pn = self.order
        staggered = self.params.oper.elem.staggered
        return pn - 2 if staggered else pn

    @property
    def order_mesh_solver(self):
        """
        Equivalent to ``lx1m``.

        .. todo:: Must include a condition to check if ALE methods are used or
                  not.

        """
        return (
            self.order
            if (
                self.params.nek.problemtype.stress_formulation
                # or ALE
            )
            else 1
        )

    @property
    def order_mhd(self):
        """Equivalent to ``lbelt``."""
        return (
            self.max_n_loc
            if "mhd" in self.params.nek.problemtype.equation.lower()
            else 1
        )

    @property
    def order_linear(self):
        """Equivalent to ``lpelt``."""
        return (
            self.max_n_loc
            if "linear" in self.params.nek.problemtype.equation.lower()
            else 1
        )

    @property
    def order_cvode(self):
        """Equivalent to ``lcvelt``."""
        return self.max_n_loc if self.params.nek.cvode._enabled else 1

    def memory_required(self):
        """According to FAQ_ the following estimate::

            lx1*ly1*lz1*lelt * 3000byte + lelg * 12byte + MPI + optional libraries
            (e.g. CVODE)

        .. _FAQ: https://nek5000.github.io/NekDoc/faq.html?highlight=memory

        """
        params = self.params
        elem = params.oper.elem

        lx1 = elem.order
        ldim = params.oper.dim
        ly1 = lx1
        lz1 = 1 + (ldim - 2) * (lx1 - 1)
        lelt = self.max_n_loc
        lelg = self.max_n_seq

        return lx1 * ly1 * lz1 * lelt * 3000 + lelg * 12

    def _str_Ln(self):
        params = self.params.oper
        str_L = map(_str_len, (params.Lx, params.Ly, params.Lz))
        str_n = map(str, (params.nx, params.ny, params.nz))
        return str_L, str_n

    def produce_str_describing_oper(self):
        """Produce a string describing the operator."""
        str_L, str_n = self._str_Ln()
        return f"{'x'.join(str_n)}_V{'x'.join(str_L)}"

    def produce_long_str_describing_oper(self, oper_method="Base"):
        """Produce a string describing the operator."""
        str_L, str_n = self._str_Ln()
        string = ""
        for key, value in zip(("Lx", "Ly", "Lz"), str_L):
            string += f"- {key} = {value}\n"
        for key, value in zip(("nx", "ny", "nz"), str_n):
            string += f"- {key} = {value}\n"
        return f"Nek5000 operator:\n{string}"

    def write_box(self, template, fp=sys.stdout, comments=""):
        """Write the .box file which is input for the ``genbox`` meshing
        tool.

        Parameters
        ----------
        template : jinja2.environment.Template
            Template instance like :code:`abl.templates.box`
        fp : io.TextIOWrapper
            File handler to write to

        """
        params = self.params
        comments += """
Autogenerated using snek5000.operators.Operators.write_box()

If dim < 0 .re2 file will be generated

If nelx (y or z) < 0, then genbox automatically generates the
                      grid spacing in the x (y or z) direction
                      with a geometric ratio given by "ratio".
                      ( ratio=1 implies uniform spacing )

Note that the character bcs _must_ have 3 spaces.
"""

        def _str_grid(*args):
            fmt = "{:.4f} {:.4f} {:.4f}"
            args = (float(value) for value in args)
            return fmt.format(*args)

        boundary = self.params.oper.boundary

        for bc in boundary:
            if len(bc) > 3:
                raise ValueError(
                    f"Length of boundary condition {bc} shall not exceed 3 characters"
                )

        # A dictionary mapping a comment to grid
        grid_info = {
            "nelx nely nelz": " ".join(
                str(-n) for n in (params.oper.nx, params.oper.ny, params.oper.nz)
            ),
            "x0 x1 ratio": _str_grid(
                params.oper.origin_x, params.oper.Lx, params.oper.ratio_x
            ),
            "y0 y1 ratio": _str_grid(
                params.oper.origin_y, params.oper.Ly, params.oper.ratio_y
            ),
            "z0 z1 ratio": _str_grid(
                params.oper.origin_z, params.oper.Lz, params.oper.ratio_z
            ),
            "BCs: (cbx0, cbx1, cby0, cby1, cbz0, cbz1)": ",".join(
                bc.ljust(3) for bc in boundary
            ),
        }
        options = {
            "comments": comments,
            "dim": str(-params.oper.dim),
            "grid_info": grid_info,
            "nb_fields": str(self.nb_fields),  # scalars + velocity
        }

        output = template.render(**options)
        fp.write(output)

    def write_size(self, template, fp=sys.stdout, comments=""):
        """Write the SIZE file which is required for compilation.

        :param jinja2.environment.Template template: Template instance like
            :code:`abl.templates.size`

        """
        comments += """
Autogenerated using snek5000.operators.Operators.write_size()
"""
        options = {"comments": comments, "params": self.params}
        # Include all @property attributes to the template
        options.update(
            {
                name: getattr(self, name)
                for name, _ in inspect.getmembers(
                    self.__class__, lambda attr: isinstance(attr, property)
                )
            }
        )
        output = template.render(**options)
        fp.write(output)


Operators.__doc__ += "\n" + docstring_params(Operators, indent_len=4)



import os
import shutil

from snek5000 import util


def test_activate_paths():
    # Remove Nek5000/bin from PATH for the sake of the test
    os.environ["PATH"] = ":".join(
        path for path in os.getenv("PATH").split(":") if "Nek5000/bin" not in path
    )
    assert shutil.which("makenek") is None, "Nek5000/bin is globally available?!"

    util.activate_paths()
    assert shutil.which("makenek"), "Nek5000/bin is not included in the PATH"



"""KTH base solver
==================

"""
from ..util import docstring_params
from .base import SimulNek


class SimulKTH(SimulNek):
    """A base class which incorporates parameters for KTH toolbox also."""

    @staticmethod
    def _complete_params_with_default(params):
        params = SimulNek._complete_params_with_default(params)

        for section in ("runpar", "monitor", "chkpoint", "stat"):
            params.nek._set_child(section)

        attribs = {"parf_write": False, "parf_name": "outparfile"}
        params.nek.runpar._set_attribs(attribs)
        params.nek.runpar._set_doc(
            """
    *Runtime parameter section for rprm module*
    :parf_write: Do we write runtime parameter file
    :parf_name: Runtime parameter file name for output (without .par)
"""
        )

        attribs = {"log_level": 4, "wall_time": "23:45"}
        params.nek.monitor._set_attribs(attribs)
        params.nek.monitor._set_doc(
            """
    *Runtime parameter section for monitor module*
    :log_level: Logging threshold for toolboxes
    :wall_time: Simulation wall time
"""
        )

        attribs = {"read_chkpt": False, "chkp_fnumber": 1, "chkp_interval": 250}
        params.nek.chkpoint._set_attribs(attribs)
        params.nek.chkpoint._set_doc(
            """
    *Runtime parameter section for checkpoint module*
    :read_chkpt: Restart from checkpoint
    :chkp_fnumber: Restart file number
    :chkp_interval: Checkpiont saving frequency (number of time steps)
"""
        )

        attribs = {"av_step": 4, "io_step": 50}
        params.nek.stat._set_attribs(attribs)
        params.nek.stat._set_doc(
            """
    *Runtime parameter section for statistics module*
    :av_step: Frequency of averaging
    :io_step: Frequency of file saving
"""
        )

        # Document all params
        for child_name in params.nek._tag_children:
            child = getattr(params.nek, child_name)
            child._autodoc_par(indent=4)

        return params


Simul = SimulKTH
Simul.__doc__ += "\n" + docstring_params(Simul, indent_len=4)



"""Runtime parameters
=====================
Scripting interface for Nek5000 :ref:`parameter file <nek:case_files_par>`.

"""
import textwrap
from configparser import ConfigParser
from io import StringIO
from math import nan
from pathlib import Path
from sys import stdout

from fluiddyn.util import import_class
from fluidsim.base.params import Parameters as _Parameters
from inflection import camelize, underscore

from .info import InfoSolverBase

literal_python2nek = {
    nan: "<real>",
    "nan": "nan",
    "None": "none",
    "True": "yes",
    "False": "no",
}
literal_nek2python = {v: k for k, v in literal_python2nek.items()}
literal_prune = ("<real>", "", "nan")


def camelcase(value):
    return camelize(str(value).lower(), uppercase_first_letter=False)


def _check_user_param(idx):
    if idx > 20:
        raise ValueError(f"userParam {idx} > 20")


class Parameters(_Parameters):
    """Container for reading, modifying and writing par_ files.

    .. _par: https://nek5000.github.io/NekDoc/problem_setup/case_files.html#parameter-file-par

    :param tag: A string representing name of case files (for example: provide
                 ``"abl"`` for files like ``abl.usr, abl.par`` etc).

    .. automethod: _read_par


    :todo: Consolidate the logic of param to par file syncing in two methods!
           More tests to see if it works.

    """

    def __init__(self, *args, **kwargs):
        comments = ("#",)
        self._set_internal_attr(
            "_par_file",
            ConfigParser(comment_prefixes=comments, inline_comment_prefixes=comments),
        )
        # Only enabled parameters would be written into par file
        self._set_internal_attr("_enabled", True)
        # User parameters sections should begin with an underscore
        self._set_internal_attr("_user", True)

        if "path_file" in kwargs:
            raise ValueError(
                "Loading directly from path_file is not supported. Use "
                "Simul.load_params_from_file() instead."
            )

        super().__init__(*args, **kwargs)

        # Like in Python Nek5000's par files are case insensitive.
        # However for consistency, case sensitivity is enforced:
        self._par_file.optionxform = str

    def _make_dict_attribs(self):
        d = super()._make_dict_attribs()
        # Append internal attributes
        d.update({"_enabled": self._enabled, "_user": self._user})
        return d

    def _read_par(self, path=None):
        """Read par file into a class children and attributes."""
        if not path:
            path = self._tag + ".par"

        self._par_file.read(path)

        for section in self._par_file.sections():
            params_child = getattr(self, section.lower().lstrip("_"))

            for option, value in self._par_file.items(section):
                if value in literal_nek2python:
                    value = literal_nek2python[value]

                # userParam%% -> user_params
                if option.lower().startswith("userparam"):
                    idx_uparam = int(option[-2:])
                    _check_user_param(idx_uparam)
                    params_child.user_params.update({idx_uparam: float(value)})
                else:
                    attrib = underscore(option)
                    setattr(params_child, attrib, value)

    def _update_par_section(self, section_name, section_dict):
        """Updates a section of the ``par_file`` object from a dictionary."""
        par = self._par_file

        # Start with underscore if it is a user section
        section_name_par = "_" if section_dict["_user"] else ""
        section_name_par += section_name.upper().lstrip("_")

        if section_name_par not in par.sections():
            par.add_section(section_name_par)
        for option, value in section_dict.items():
            # Convert to string to avoid hash collisions
            # hash(1) == hash(True)
            literal = str(value) if value is not nan else nan
            if literal in literal_python2nek:
                value = literal_python2nek[literal]

            if value in literal_prune:
                continue

            # Make everything consistent where values refer to option names
            # if option in ("stop_at", "write_control"):
            if str(value) in section_dict:
                value = camelcase(value)

            # user_params -> userParam%%
            if option.lower().startswith("user_params") and isinstance(value, dict):
                for idx_uparam, value_uparam in value.items():
                    _check_user_param(idx_uparam)
                    par.set(
                        section_name_par,
                        f"userParam{idx_uparam:02d}",
                        str(value_uparam),
                    )
            else:
                par.set(section_name_par, camelcase(option), str(value))

    def _sync_par(self):
        """Sync values in param children and attributes to ``self._par_file``
        object.

        """
        if self._tag_children:
            data = [
                (child, getattr(self, child)._make_dict_tree())
                for child in self._tag_children
            ]
        else:
            # No children
            data = [(self._tag, self._make_dict_attribs())]

        for child, d in data:
            section_name = child.upper()
            self._update_par_section(section_name, d)

        self._tidy_par()

    def _tidy_par(self):
        """Remove internal attributes and disabled sections from par file."""
        par = self._par_file
        for section_name in par.sections():
            par.remove_option(section_name, "_user")

            enabled = par.getboolean(section_name, "_enabled")
            if enabled:
                par.remove_option(section_name, "_enabled")
            else:
                par.remove_section(section_name)

    def _write_par(self, path=stdout):
        """Write contents of ``self._par_file`` to file handler opened in disk
        or to stdout.

        """
        self._sync_par()
        if isinstance(path, (str, Path)):
            with open(path, "w") as fp:
                self._par_file.write(fp)
        else:
            self._par_file.write(path)

    def _autodoc_par(self, indent=0):
        """Autodoc a code block with ``ini`` syntax and set docstring."""
        self._sync_par()
        docstring = "\n.. code-block:: ini\n\n"
        with StringIO() as output:
            self._par_file.write(output)
            ini = output.getvalue()

        docstring += textwrap.indent(ini, "   ")

        if ini:
            self._set_doc(self._doc + textwrap.indent(docstring, " " * indent))


def create_params(input_info_solver):
    """Create a Parameters instance from an InfoSolverBase instance."""
    if isinstance(input_info_solver, InfoSolverBase):
        info_solver = input_info_solver
    elif hasattr(input_info_solver, "Simul"):
        info_solver = input_info_solver.Simul.create_default_params()
    else:
        raise ValueError("Can not create params from input input_info_solver.")

    params = Parameters(tag="params")
    dict_classes = info_solver.import_classes()

    dict_classes["Solver"] = import_class(
        info_solver.module_name, info_solver.class_name
    )

    for Class in list(dict_classes.values()):
        if hasattr(Class, "_complete_params_with_default"):
            try:
                Class._complete_params_with_default(params)
            except TypeError:
                try:
                    Class._complete_params_with_default(params, info_solver)
                except TypeError as e:
                    e.args += ("for class: " + repr(Class),)
                    raise

    return params



"""Barebones setup.py retained for supporting editable_ installs::

    pip install -e .

.. _editable: https://github.com/pypa/packaging-problems/issues/256#issuecomment-504790004

"""
from setuptools import setup

setup()



import tempfile
from io import StringIO
from pathlib import Path

from snek5000.log import logger
from snek5000.params import Parameters
from snek5000.util import init_params


def test_empty_params():
    params = Parameters(tag="empty")
    params._write_par()


def test_simul_params():
    from snek5000.solvers.base import Simul

    params = Simul.create_default_params()

    buffer = StringIO()
    params.nek._write_par(buffer)
    params.nek._read_par(buffer)
    buffer2 = StringIO()
    params.nek._write_par(buffer2)

    assert buffer.getvalue() == buffer2.getvalue()


def test_oper_params(oper):
    from snek5000.operators import Operators

    params = init_params(Operators)
    logger.debug(params.oper.max)
    logger.debug(params.oper.max._doc)
    logger.debug(params.oper.elem)
    logger.debug(params.oper.elem._doc)


def test_par_xml_match():
    from phill.solver import Simul

    params = Simul.create_default_params()
    output1 = StringIO()
    params.nek._write_par(output1)

    tmp_dir = Path(tempfile.mkdtemp("snek5000", __name__))
    params_xml = params._save_as_xml(str(tmp_dir / "params.xml"))

    try:
        from snek5000.params import Parameters

        nparams = Parameters(tag="params", path_file=params_xml)
    except ValueError:
        # Should raise an error
        pass
    else:
        raise ValueError("Parameters(path_file=...) worked unexpectedly.")

    nparams = Simul.load_params_from_file(path_xml=params_xml)
    output2 = StringIO()
    nparams.nek._write_par(output2)

    par1 = output1.getvalue()
    par2 = output2.getvalue()
    output1.close()
    output2.close()

    def format_sections(params):
        par = params.nek._par_file

        # no options in the section
        for section_name in par.sections():
            if not par.options(section_name):
                par.remove_section(section_name)

        return sorted(par.sections())

    assert format_sections(params) == format_sections(nparams)

    def format_par(text):
        """Sort non-blank lines"""
        from ast import literal_eval

        ftext = []
        for line in text.splitlines():
            # not blank
            if line:
                # Uniform format for literals
                if " = " in line:
                    key, value = line.split(" = ")
                    try:
                        line = " = ".join([key, str(literal_eval(value))])
                    except (SyntaxError, ValueError):
                        pass

                ftext.append(line)

        return sorted(ftext)

    assert format_par(par1) == format_par(par2)



"""Logging
==========
A logger instance (variable :code:`logger`).

"""

import logging
import os

logger = logging.getLogger(__name__)


# Initialize logging
try:
    # Set a nice colored output
    from colorlog import ColoredFormatter

    formatter = ColoredFormatter(
        "%(log_color)s%(levelname)-8s%(reset)s %(blue)s%(message)s",
        log_colors={
            "DEBUG": "cyan",
            "INFO": "green",
            "WARNING": "yellow",
            "ERROR": "red",
            "CRITICAL": "red",
        },
    )
    stream = logging.StreamHandler()
    stream.setFormatter(formatter)
    logger.addHandler(stream)
except ImportError:
    # No color available, use default config
    logging.basicConfig(format="%(levelname)s: %(message)s")
    logger.info("Disabling color, you really want to install colorlog.")


if os.getenv("SNEK_DEBUG"):
    level = logging.DEBUG
else:
    level = logging.INFO


logger.setLevel(level)
